# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendOptional,
)
from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import LegendApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SortItem,
    SortItemOrdering,
    SortItemNullOrdering,
    SingleColumn,
    Expression,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_base_tds_frame import LegendApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "SortFunction"
]


class SortFunction(LegendApiAppliedFunction):
    __base_frame: LegendApiBaseTdsFrame
    __column_name_list: PyLegendList[str]
    __directions: PyLegendOptional[PyLegendList[str]]

    @classmethod
    def name(cls) -> str:
        return "sort"

    def __init__(
            self,
            base_frame: LegendApiBaseTdsFrame,
            column_name_list: PyLegendList[str],
            directions: PyLegendOptional[PyLegendList[str]]
    ) -> None:
        self.__base_frame = base_frame
        self.__column_name_list = column_name_list
        self.__directions = directions

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        db_extension = config.sql_to_string_generator().get_db_extension()

        def find_column_expression(col: str, query: QuerySpecification) -> Expression:
            filtered = [
                s for s in query.select.selectItems
                if (isinstance(s, SingleColumn) and
                    s.alias == db_extension.quote_identifier(col))
            ]
            if len(filtered) == 0:
                raise RuntimeError("Cannot find column: " + col)  # pragma: no cover
            return filtered[0].expression

        should_create_sub_query = (base_query.offset is not None) or (base_query.limit is not None)
        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )

        direction_list = self._build_directions_list()

        new_query.orderBy = [
            SortItem(
                sortKey=find_column_expression(self.__column_name_list[i], new_query),
                ordering=(SortItemOrdering.ASCENDING if direction_list[i].upper() == "ASC"
                          else SortItemOrdering.DESCENDING),
                nullOrdering=SortItemNullOrdering.UNDEFINED
            )
            for i in range(len(self.__column_name_list))
        ]
        return new_query

    def base_frame(self) -> LegendApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        base_cols = [c.get_name() for c in self.__base_frame.columns()]
        for c in self.__column_name_list:
            if c not in base_cols:
                raise ValueError(
                    f"Column - '{c}' in sort columns list doesn't exist in the current frame. "
                    f"Current frame columns: {base_cols}"
                )

        if (self.__directions is not None) and (len(self.__directions) > 0):
            if len(self.__column_name_list) != len(self.__directions):
                cols = self.__column_name_list
                dirs = self.__directions
                raise ValueError(
                    "Sort directions (ASC/DESC) provided need to be in sync with columns or left empty to "
                    f"choose defaults. Passed column list: {cols}, directions: {dirs}"
                )

            for d in self.__directions:
                if d.upper() not in (["ASC", "DESC"]):
                    raise ValueError(
                        "Sort direction can be ASC/DESC (case insensitive). Passed unknown value: " + d
                    )

        return True

    def _build_directions_list(self) -> PyLegendList[str]:
        if (self.__directions is None) or (len(self.__directions) == 0):
            return ["ASC" for _ in self.__column_name_list]
        else:
            # Already validated that directions are all ASC/DESC and length matches with column list
            direction_list = []
            for d in self.__directions:
                direction_list.append(d.upper())
            return direction_list



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import date, datetime
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
)
from pylegend.core.language.expression import (
    PyLegendExpression,
    PyLegendExpressionBooleanReturn,
    PyLegendExpressionStringReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionDateTimeReturn,
    PyLegendExpressionStrictDateReturn,
)
from pylegend.core.sql.metamodel import (
    Expression,
    BooleanLiteral,
    StringLiteral,
    IntegerLiteral,
    DoubleLiteral,
    QuerySpecification,
    Cast,
    ColumnType,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig


__all__: PyLegendSequence[str] = [
    "PyLegendBooleanLiteralExpression",
    "PyLegendStringLiteralExpression",
    "PyLegendIntegerLiteralExpression",
    "PyLegendFloatLiteralExpression",
    "PyLegendDateTimeLiteralExpression",
    "PyLegendStrictDateLiteralExpression",
    "convert_literal_to_literal_expression",
]


class PyLegendBooleanLiteralExpression(PyLegendExpressionBooleanReturn):
    __value: bool

    def __init__(self, value: bool) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return BooleanLiteral(value=self.__value)


class PyLegendStringLiteralExpression(PyLegendExpressionStringReturn):
    __value: str

    def __init__(self, value: str) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringLiteral(value=self.__value, quoted=False)


class PyLegendIntegerLiteralExpression(PyLegendExpressionIntegerReturn):
    __value: int

    def __init__(self, value: int) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return IntegerLiteral(value=self.__value)


class PyLegendFloatLiteralExpression(PyLegendExpressionFloatReturn):
    __value: float

    def __init__(self, value: float) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return DoubleLiteral(value=self.__value)


class PyLegendDateTimeLiteralExpression(PyLegendExpressionDateTimeReturn):
    __value: datetime

    def __init__(self, value: datetime) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return Cast(
            expression=StringLiteral(value=self.__value.isoformat(), quoted=False),
            type_=ColumnType(name="TIMESTAMP", parameters=[])
        )


class PyLegendStrictDateLiteralExpression(PyLegendExpressionStrictDateReturn):
    __value: date

    def __init__(self, value: date) -> None:
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return Cast(
            expression=StringLiteral(value=self.__value.isoformat(), quoted=False),
            type_=ColumnType(name="DATE", parameters=[])
        )


def convert_literal_to_literal_expression(
        literal: PyLegendUnion[int, float, bool, str, datetime, date]
) -> PyLegendExpression:
    if isinstance(literal, bool):
        return PyLegendBooleanLiteralExpression(literal)
    if isinstance(literal, int):
        return PyLegendIntegerLiteralExpression(literal)
    if isinstance(literal, float):
        return PyLegendFloatLiteralExpression(literal)
    if isinstance(literal, str):
        return PyLegendStringLiteralExpression(literal)
    if isinstance(literal, datetime):
        return PyLegendDateTimeLiteralExpression(literal)
    if isinstance(literal, date):
        return PyLegendStrictDateLiteralExpression(literal)

    raise TypeError(f"Cannot convert value - {literal} of type {type(literal)} to literal expression")



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from datetime import date, datetime
from pylegend._typing import (
    PyLegendSequence,
    PyLegendUnion,
)
from pylegend.core.language.primitives.primitive import PyLegendPrimitiveOrPythonPrimitive
from pylegend.core.language import (
    PyLegendInteger,
    PyLegendFloat,
    PyLegendNumber,
    PyLegendString,
    PyLegendBoolean,
    PyLegendDate,
    PyLegendDateTime,
    PyLegendStrictDate,
    convert_literal_to_literal_expression,
)
from pylegend.core.language.operations.collection_operation_expressions import (
    PyLegendCountExpression,
    PyLegendDistinctCountExpression,
    PyLegendAverageExpression,
    PyLegendIntegerMaxExpression,
    PyLegendIntegerMinExpression,
    PyLegendIntegerSumExpression,
    PyLegendFloatMaxExpression,
    PyLegendFloatMinExpression,
    PyLegendFloatSumExpression,
    PyLegendNumberMaxExpression,
    PyLegendNumberMinExpression,
    PyLegendNumberSumExpression,
    PyLegendStdDevSampleExpression,
    PyLegendStdDevPopulationExpression,
    PyLegendVarianceSampleExpression,
    PyLegendVariancePopulationExpression,
    PyLegendStringMaxExpression,
    PyLegendStringMinExpression,
    PyLegendJoinStringsExpression,
    PyLegendStrictDateMaxExpression,
    PyLegendStrictDateMinExpression,
    PyLegendDateMaxExpression,
    PyLegendDateMinExpression,
)


__all__: PyLegendSequence[str] = [
    "PyLegendPrimitiveCollection",
    "PyLegendIntegerCollection",
    "PyLegendFloatCollection",
    "PyLegendNumberCollection",
    "PyLegendStringCollection",
    "PyLegendBooleanCollection",
    "PyLegendDateCollection",
    "PyLegendDateTimeCollection",
    "PyLegendStrictDateCollection",
    "create_primitive_collection"
]


class PyLegendPrimitiveCollection(metaclass=ABCMeta):
    __nested: PyLegendPrimitiveOrPythonPrimitive

    def __init__(self, nested: PyLegendPrimitiveOrPythonPrimitive) -> None:
        self.__nested = nested

    def count(self) -> "PyLegendInteger":
        if isinstance(self.__nested, (bool, int, float, str, date, datetime)):
            nested_expr = convert_literal_to_literal_expression(self.__nested)
        else:
            nested_expr = self.__nested.value()
        return PyLegendInteger(PyLegendCountExpression(nested_expr))

    def distinct_count(self) -> "PyLegendInteger":
        if isinstance(self.__nested, (bool, int, float, str, date, datetime)):
            nested_expr = convert_literal_to_literal_expression(self.__nested)
        else:
            nested_expr = self.__nested.value()
        return PyLegendInteger(PyLegendDistinctCountExpression(nested_expr))


class PyLegendNumberCollection(PyLegendPrimitiveCollection):
    __nested: PyLegendUnion[int, float, PyLegendInteger, PyLegendFloat, PyLegendNumber]

    def __init__(self, nested: PyLegendUnion[int, float, PyLegendInteger, PyLegendFloat, PyLegendNumber]) -> None:
        super().__init__(nested)
        self.__nested = nested

    def average(self) -> "PyLegendFloat":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendFloat(PyLegendAverageExpression(nested_expr))  # type: ignore

    def max(self) -> "PyLegendNumber":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendNumber(PyLegendNumberMaxExpression(nested_expr))  # type: ignore

    def min(self) -> "PyLegendNumber":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendNumber(PyLegendNumberMinExpression(nested_expr))  # type: ignore

    def sum(self) -> "PyLegendNumber":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendNumber(PyLegendNumberSumExpression(nested_expr))  # type: ignore

    def std_dev_sample(self) -> "PyLegendNumber":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendNumber(PyLegendStdDevSampleExpression(nested_expr))  # type: ignore

    def std_dev(self) -> "PyLegendNumber":
        return self.std_dev_sample()

    def std_dev_population(self) -> "PyLegendNumber":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendNumber(PyLegendStdDevPopulationExpression(nested_expr))  # type: ignore

    def variance_sample(self) -> "PyLegendNumber":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendNumber(PyLegendVarianceSampleExpression(nested_expr))  # type: ignore

    def variance(self) -> "PyLegendNumber":
        return self.variance_sample()

    def variance_population(self) -> "PyLegendNumber":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (int, float))
            else self.__nested.value()
        )
        return PyLegendNumber(PyLegendVariancePopulationExpression(nested_expr))  # type: ignore


class PyLegendIntegerCollection(PyLegendNumberCollection):
    __nested: PyLegendUnion[int, PyLegendInteger]

    def __init__(self, nested: PyLegendUnion[int, PyLegendInteger]) -> None:
        super().__init__(nested)
        self.__nested = nested

    def max(self) -> "PyLegendInteger":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, int)
            else self.__nested.value()
        )
        return PyLegendInteger(PyLegendIntegerMaxExpression(nested_expr))  # type: ignore

    def min(self) -> "PyLegendInteger":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, int)
            else self.__nested.value()
        )
        return PyLegendInteger(PyLegendIntegerMinExpression(nested_expr))  # type: ignore

    def sum(self) -> "PyLegendInteger":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, int)
            else self.__nested.value()
        )
        return PyLegendInteger(PyLegendIntegerSumExpression(nested_expr))  # type: ignore


class PyLegendFloatCollection(PyLegendNumberCollection):
    __nested: PyLegendUnion[float, PyLegendFloat]

    def __init__(self, nested: PyLegendUnion[float, PyLegendFloat]) -> None:
        super().__init__(nested)
        self.__nested = nested

    def max(self) -> "PyLegendFloat":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, float)
            else self.__nested.value()
        )
        return PyLegendFloat(PyLegendFloatMaxExpression(nested_expr))  # type: ignore

    def min(self) -> "PyLegendFloat":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, float)
            else self.__nested.value()
        )
        return PyLegendFloat(PyLegendFloatMinExpression(nested_expr))  # type: ignore

    def sum(self) -> "PyLegendFloat":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, float)
            else self.__nested.value()
        )
        return PyLegendFloat(PyLegendFloatSumExpression(nested_expr))  # type: ignore


class PyLegendStringCollection(PyLegendPrimitiveCollection):
    __nested: PyLegendUnion[str, PyLegendString]

    def __init__(self, nested: PyLegendUnion[str, PyLegendString]) -> None:
        super().__init__(nested)
        self.__nested = nested

    def max(self) -> "PyLegendString":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, str)
            else self.__nested.value()
        )
        return PyLegendString(PyLegendStringMaxExpression(nested_expr))  # type: ignore

    def min(self) -> "PyLegendString":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, str)
            else self.__nested.value()
        )
        return PyLegendString(PyLegendStringMinExpression(nested_expr))  # type: ignore

    def join(self, separator: str) -> "PyLegendString":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, str)
            else self.__nested.value()
        )
        separator_expr = convert_literal_to_literal_expression(separator)
        return PyLegendString(PyLegendJoinStringsExpression(nested_expr, separator_expr))  # type: ignore


class PyLegendBooleanCollection(PyLegendPrimitiveCollection):
    __nested: PyLegendUnion[bool, PyLegendBoolean]

    def __init__(self, nested: PyLegendUnion[bool, PyLegendBoolean]) -> None:
        super().__init__(nested)
        self.__nested = nested


class PyLegendDateCollection(PyLegendPrimitiveCollection):
    __nested: PyLegendUnion[date, datetime, PyLegendDate]

    def __init__(self, nested: PyLegendUnion[date, datetime, PyLegendDate]) -> None:
        super().__init__(nested)
        self.__nested = nested

    def max(self) -> "PyLegendDate":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (date, datetime))
            else self.__nested.value()
        )
        return PyLegendDate(PyLegendDateMaxExpression(nested_expr))  # type: ignore

    def min(self) -> "PyLegendDate":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, (date, datetime))
            else self.__nested.value()
        )
        return PyLegendDate(PyLegendDateMinExpression(nested_expr))  # type: ignore


class PyLegendDateTimeCollection(PyLegendDateCollection):
    __nested: PyLegendUnion[datetime, PyLegendDateTime]

    def __init__(self, nested: PyLegendUnion[datetime, PyLegendDateTime]) -> None:
        super().__init__(nested)
        self.__nested = nested


class PyLegendStrictDateCollection(PyLegendDateCollection):
    __nested: PyLegendUnion[date, PyLegendStrictDate]

    def __init__(self, nested: PyLegendUnion[date, PyLegendStrictDate]) -> None:
        super().__init__(nested)
        self.__nested = nested

    def max(self) -> "PyLegendStrictDate":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, date)
            else self.__nested.value()
        )
        return PyLegendStrictDate(PyLegendStrictDateMaxExpression(nested_expr))  # type: ignore

    def min(self) -> "PyLegendStrictDate":
        nested_expr = (
            convert_literal_to_literal_expression(self.__nested) if isinstance(self.__nested, date)
            else self.__nested.value()
        )
        return PyLegendStrictDate(PyLegendStrictDateMinExpression(nested_expr))  # type: ignore


def create_primitive_collection(nested: PyLegendPrimitiveOrPythonPrimitive) -> PyLegendPrimitiveCollection:
    if isinstance(nested, (int, PyLegendInteger)):
        return PyLegendIntegerCollection(nested)

    if isinstance(nested, (float, PyLegendFloat)):
        return PyLegendFloatCollection(nested)

    if isinstance(nested, PyLegendNumber):
        return PyLegendNumberCollection(nested)

    if isinstance(nested, (str, PyLegendString)):
        return PyLegendStringCollection(nested)

    if isinstance(nested, (bool, PyLegendBoolean)):
        return PyLegendBooleanCollection(nested)

    if isinstance(nested, (datetime, PyLegendDateTime)):
        return PyLegendDateTimeCollection(nested)

    if isinstance(nested, (date, PyLegendStrictDate)):
        return PyLegendStrictDateCollection(nested)

    if isinstance(nested, PyLegendDate):
        return PyLegendDateCollection(nested)

    raise RuntimeError(f"Not supported type - {type(nested)}")  # pragma: no cover



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendTuple
)
from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import LegendApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
    SelectItem
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_base_tds_frame import LegendApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "RestrictFunction"
]


class RestrictFunction(LegendApiAppliedFunction):
    __base_frame: LegendApiBaseTdsFrame
    __column_name_list: PyLegendList[str]

    @classmethod
    def name(cls) -> str:
        return "restrict"

    def __init__(self, base_frame: LegendApiBaseTdsFrame, column_name_list: PyLegendList[str]) -> None:
        self.__base_frame = base_frame
        self.__column_name_list = column_name_list

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        db_extension = config.sql_to_string_generator().get_db_extension()
        columns_to_retain = [db_extension.quote_identifier(x) for x in self.__column_name_list]

        sub_query_required = (len(base_query.groupBy) > 0) or (len(base_query.orderBy) > 0) or \
                             (base_query.having is not None) or base_query.select.distinct

        if sub_query_required:
            new_query = create_sub_query(base_query, config, "root", columns_to_retain=columns_to_retain)
            return new_query
        else:
            new_cols_with_index: PyLegendList[PyLegendTuple[int, 'SelectItem']] = []
            for col in base_query.select.selectItems:
                if not isinstance(col, SingleColumn):
                    raise ValueError("Restrict operation not supported for queries "
                                     "with columns other than SingleColumn")  # pragma: no cover
                if col.alias is None:
                    raise ValueError("Restrict operation not supported for queries "
                                     "with SingleColumns with missing alias")  # pragma: no cover
                if col.alias in columns_to_retain:
                    new_cols_with_index.append((columns_to_retain.index(col.alias), col))

            new_select_items = [y[1] for y in sorted(new_cols_with_index, key=lambda x: x[0])]
            new_query = copy_query(base_query)
            new_query.select.selectItems = new_select_items
            return new_query

    def base_frame(self) -> LegendApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        base_columns = self.__base_frame.columns()
        new_columns = []
        for c in self.__column_name_list:
            for base_col in base_columns:
                if base_col.get_name() == c:
                    new_columns.append(base_col.copy())
                    break
        return new_columns

    def validate(self) -> bool:
        base_columns = self.__base_frame.columns()
        for c in self.__column_name_list:
            found_col = False
            for base_col in base_columns:
                if base_col.get_name() == c:
                    found_col = True
                    break
            if not found_col:
                raise ValueError(
                    f"Column - '{c}' in restrict columns list doesn't exist in the current frame. "
                    f"Current frame columns: {[x.get_name() for x in base_columns]}"
                )
        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ijson  # type: ignore
import pandas as pd
import numpy as np
from pylegend._typing import (
    PyLegendSequence,
    TYPE_CHECKING
)
if TYPE_CHECKING:
    from pylegend.core.tds.tds_frame import PyLegendTdsFrame  # pragma: no cover
from pylegend.core.tds.tds_column import TdsColumn, PrimitiveTdsColumn
from pylegend.core.tds.result_handler.result_handler import ResultHandler
from pylegend.core.request.response_reader import ResponseReader

__all__: PyLegendSequence[str] = [
    "ToPandasDfResultHandler",
    "PandasDfReadConfig"
]

DATE_TYPES = ["StrictDate", "DateTime", "Date"]

COLUMN_TYPE_DTYPE_MAP = {
    "Boolean": "boolean",
    "Integer": "Int64",
    "Float": "Float64",
    "Number": "Float64",
    "String": "object"
}


class PandasDfReadConfig:
    __parse_float_as_decimal: bool
    __rows_per_batch: int

    def __init__(self, parse_float_as_decimal: bool = False, rows_per_batch: int = 16_384) -> None:
        self.__parse_float_as_decimal = parse_float_as_decimal
        self.__rows_per_batch = rows_per_batch

    def rows_per_batch(self) -> int:
        return self.__rows_per_batch

    def parse_float_as_decimal(self) -> bool:
        return self.__parse_float_as_decimal


class ToPandasDfResultHandler(ResultHandler[pd.DataFrame]):
    __pandas_df_read_config: PandasDfReadConfig

    def __init__(
            self,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> None:
        self.__pandas_df_read_config = pandas_df_read_config

    def handle_result(self, frame: "PyLegendTdsFrame", result: ResponseReader) -> pd.DataFrame:
        df: pd.DataFrame = pd.concat(
            self._read_partial_dfs(
                frame,
                ijson.items(
                    result,
                    "result.rows.item.values",
                    use_float=not self.__pandas_df_read_config.parse_float_as_decimal()
                ),
            ),
            ignore_index=True
        )
        return df

    def _read_partial_dfs(self, frame: "PyLegendTdsFrame", row_iter):  # type: ignore
        all_values_list = []
        cnt = 0
        for row in row_iter:
            all_values_list.extend(row)
            cnt += 1
            if cnt == self.__pandas_df_read_config.rows_per_batch():
                yield ToPandasDfResultHandler._create_df_from_list(all_values_list, frame)  # type: ignore
                all_values_list = []
                cnt = 0

        if cnt > 0:
            yield ToPandasDfResultHandler._create_df_from_list(all_values_list, frame)  # type: ignore
            all_values_list = []
            cnt = 0

        if cnt == 0 and len(all_values_list) == 0:
            return
        else:
            raise RuntimeError("Unexpected state")  # pragma: no cover

    @staticmethod
    def _create_df_from_list(all_values_list, frame):  # type: ignore
        columns = frame.columns()
        columns_length = len(columns)
        column_series_list = [
            ToPandasDfResultHandler._create_series(all_values_list, x, i, columns_length)
            for (i, x) in enumerate(columns)
        ]
        return pd.concat(column_series_list, axis=1)

    @staticmethod
    def _create_series(all_values_list, column: TdsColumn, col_index: int, columns_length: int):  # type: ignore
        if isinstance(column, PrimitiveTdsColumn):
            if column.get_type() in DATE_TYPES:
                dtype = None
            elif column.get_type() in COLUMN_TYPE_DTYPE_MAP:
                dtype = COLUMN_TYPE_DTYPE_MAP[column.get_type()]
            else:
                raise RuntimeError(  # pragma: no cover
                    f"Cannot infer pandas column dtype for column '{column.get_name()}' with type '{column.get_type()}'"
                )
        else:
            dtype = "object"

        if dtype:
            series = pd.Series(
                data=all_values_list[col_index::columns_length],
                name=column.get_name(),
                dtype=dtype
            )
        else:
            series = pd.Series(
                data=all_values_list[col_index::columns_length],
                name=column.get_name()
            )

        if dtype == "object":
            series = series.fillna(np.nan)

        if column.get_type() in DATE_TYPES:
            series = pd.to_datetime(series, format="ISO8601")

        return series



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import ijson  # type: ignore
from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    TYPE_CHECKING
)
if TYPE_CHECKING:
    from pylegend.core.tds.tds_frame import PyLegendTdsFrame  # pragma: no cover
from pylegend.core.tds.result_handler.result_handler import ResultHandler
from pylegend.core.request.response_reader import ResponseReader

__all__: PyLegendSequence[str] = [
    "ToCsvFileResultHandler"
]


class ToCsvFileResultHandler(ResultHandler[None]):

    def __init__(self, file_or_writer, parse_float_as_decimal: bool = False) -> None:  # type: ignore
        self.__file: PyLegendOptional[str] = file_or_writer if isinstance(file_or_writer, str) else None
        self.__csv_writer = None if isinstance(file_or_writer, str) else file_or_writer
        self.__parse_float_as_decimal = parse_float_as_decimal

    def handle_result(self, frame: "PyLegendTdsFrame", result: ResponseReader) -> None:
        if self.__file is None:
            self.__write_result(self.__csv_writer, frame, result, self.__parse_float_as_decimal)
        else:
            with open(self.__file, 'w', newline='') as csvfile:
                csv_writer = csv.writer(csvfile)
                self.__write_result(csv_writer, frame, result, self.__parse_float_as_decimal)

    @staticmethod
    def __write_result(  # type: ignore
            csv_writer,
            frame,
            result,
            parse_as_decimal
    ) -> None:
        csv_writer.writerow([col.get_name() for col in frame.columns()])
        for next_row in ijson.items(result, "result.rows.item.values", use_float=not parse_as_decimal):
            csv_writer.writerow(next_row)
        return



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendCallable,
)
from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import LegendApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_base_tds_frame import LegendApiBaseTdsFrame
from pylegend.core.language import (
    TdsRow,
    PyLegendPrimitive,
    PyLegendPrimitiveOrPythonPrimitive,
    convert_literal_to_literal_expression,
)
from pylegend.core.tds.legend_api.frames.functions.function_helpers import tds_column_for_primitive

__all__: PyLegendSequence[str] = [
    "ExtendFunction"
]


class ExtendFunction(LegendApiAppliedFunction):
    __base_frame: LegendApiBaseTdsFrame
    __functions_list: PyLegendList[PyLegendCallable[[TdsRow], PyLegendPrimitiveOrPythonPrimitive]]
    __column_names_list: PyLegendList[str]

    @classmethod
    def name(cls) -> str:
        return "extend"

    def __init__(
            self,
            base_frame: LegendApiBaseTdsFrame,
            functions_list: PyLegendList[PyLegendCallable[[TdsRow], PyLegendPrimitiveOrPythonPrimitive]],
            column_names_list: PyLegendList[str]
    ) -> None:
        self.__base_frame = base_frame
        self.__functions_list = functions_list
        self.__column_names_list = column_names_list

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = len(base_query.groupBy) > 0
        db_extension = config.sql_to_string_generator().get_db_extension()

        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )

        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        for (func, name) in zip(self.__functions_list, self.__column_names_list):
            col_expr = func(tds_row)

            if isinstance(col_expr, (bool, int, float, str, date, datetime)):
                col_sql_expr = convert_literal_to_literal_expression(col_expr).to_sql_expression(
                    {"frame": new_query},
                    config
                )
            else:
                col_sql_expr = col_expr.to_sql_expression({"frame": new_query}, config)

            new_query.select.selectItems.append(
                SingleColumn(alias=db_extension.quote_identifier(name), expression=col_sql_expr)
            )

        return new_query

    def base_frame(self) -> LegendApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        new_columns = [c.copy() for c in self.__base_frame.columns()]

        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        for (func, name) in zip(self.__functions_list, self.__column_names_list):
            result = func(tds_row)
            new_columns.append(tds_column_for_primitive(name, result))

        return new_columns

    def validate(self) -> bool:
        if len(self.__functions_list) != len(self.__column_names_list):
            raise ValueError(
                "For extend function, function list and column names list arguments should be of same size. "
                f"Passed param sizes -  Functions: {len(self.__functions_list)}, "
                f"Column names: {len(self.__column_names_list)}"
            )

        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        index = 0
        for (func, name) in zip(self.__functions_list, self.__column_names_list):
            copy = func  # For MyPy
            if not isinstance(copy, type(lambda x: 0)) or (copy.__code__.co_argcount != 1):
                raise TypeError(
                    f"Error at extend function at index {index} (0-indexed). "
                    "Each extend function should be a lambda which takes one argument (TDSRow)"
                )
            if not isinstance(name, str):
                raise TypeError(
                    f"Error at extend column name at index {index} (0-indexed). "
                    "Column name should be a string"
                )

            try:
                result = func(tds_row)
            except Exception as e:
                raise RuntimeError(
                    f"Extend function at index {index} (0-indexed) incompatible. "
                    f"Error occurred while evaluating. Message: {str(e)}"
                ) from e

            if not isinstance(result, (int, float, bool, str, PyLegendPrimitive)):
                raise ValueError(
                    f"Extend function at index {index} (0-indexed) incompatible. "
                    f"Returns non-primitive - {str(type(result))}"
                )

            index += 1

        if len(self.__column_names_list) != len(set(self.__column_names_list)):
            raise ValueError(f"Extend column names list has duplicates: {self.__column_names_list}")

        base_cols = [c.get_name() for c in self.__base_frame.columns()]
        for c in self.__column_names_list:
            if c in base_cols:
                raise ValueError(f"Extend column name - '{c}' already exists in base frame")

        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendCallable,
)
from pylegend.core.language.expression import (
    PyLegendExpression,
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig


__all__: PyLegendSequence[str] = [
    "PyLegendBinaryExpression"
]


class PyLegendBinaryExpression(PyLegendExpression, metaclass=ABCMeta):
    __operand1: PyLegendExpression
    __operand2: PyLegendExpression
    __to_sql_func: PyLegendCallable[
        [Expression, Expression, PyLegendDict[str, QuerySpecification], FrameToSqlConfig],
        Expression
    ]

    def __init__(
            self,
            operand1: PyLegendExpression,
            operand2: PyLegendExpression,
            to_sql_func: PyLegendCallable[
                [Expression, Expression, PyLegendDict[str, QuerySpecification], FrameToSqlConfig],
                Expression
            ]
    ) -> None:
        self.__operand1 = operand1
        self.__operand2 = operand2
        self.__to_sql_func = to_sql_func

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        op1_expr = self.__operand1.to_sql_expression(frame_name_to_base_query_map, config)
        op2_expr = self.__operand2.to_sql_expression(frame_name_to_base_query_map, config)
        return self.__to_sql_func(
            op1_expr,
            op2_expr,
            frame_name_to_base_query_map,
            config
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.expression import (
    PyLegendExpressionFloatReturn,
)
from pylegend.core.language.operations.binary_expression import PyLegendBinaryExpression
from pylegend.core.language.operations.unary_expression import PyLegendUnaryExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
    ArithmeticType,
    ArithmeticExpression,
    NegativeExpression,
)
from pylegend.core.sql.metamodel_extension import (
    AbsoluteExpression,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig


__all__: PyLegendSequence[str] = [
    "PyLegendFloatAddExpression",
    "PyLegendFloatAbsoluteExpression",
    "PyLegendFloatNegativeExpression",
    "PyLegendFloatSubtractExpression",
    "PyLegendFloatMultiplyExpression",
]


class PyLegendFloatAddExpression(PyLegendBinaryExpression, PyLegendExpressionFloatReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.ADD, expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionFloatReturn, operand2: PyLegendExpressionFloatReturn) -> None:
        PyLegendExpressionFloatReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendFloatAddExpression.__to_sql_func
        )


class PyLegendFloatSubtractExpression(PyLegendBinaryExpression, PyLegendExpressionFloatReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.SUBTRACT, expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionFloatReturn, operand2: PyLegendExpressionFloatReturn) -> None:
        PyLegendExpressionFloatReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendFloatSubtractExpression.__to_sql_func
        )


class PyLegendFloatMultiplyExpression(PyLegendBinaryExpression, PyLegendExpressionFloatReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.MULTIPLY, expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionFloatReturn, operand2: PyLegendExpressionFloatReturn) -> None:
        PyLegendExpressionFloatReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendFloatMultiplyExpression.__to_sql_func
        )


class PyLegendFloatAbsoluteExpression(PyLegendUnaryExpression, PyLegendExpressionFloatReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return AbsoluteExpression(expression)

    def __init__(self, operand: PyLegendExpressionFloatReturn) -> None:
        PyLegendExpressionFloatReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendFloatAbsoluteExpression.__to_sql_func
        )


class PyLegendFloatNegativeExpression(PyLegendUnaryExpression, PyLegendExpressionFloatReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return NegativeExpression(expression)

    def __init__(self, operand: PyLegendExpressionFloatReturn) -> None:
        PyLegendExpressionFloatReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendFloatNegativeExpression.__to_sql_func
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendDict,
    PyLegendCallable,
    PyLegendUnion,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.tds_column import TdsColumn, PrimitiveTdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.language import (
    TdsRow,
    PyLegendPrimitive,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendNumber,
    PyLegendBoolean,
    PyLegendString,
)


class AssignFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __col_definitions: PyLegendDict[
        str,
        PyLegendCallable[[TdsRow], PyLegendUnion[int, float, bool, str, date, datetime, PyLegendPrimitive]],
    ]

    @classmethod
    def name(cls) -> str:
        return "assign"

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            col_definitions: PyLegendDict[
                str,
                PyLegendCallable[[TdsRow], PyLegendUnion[int, float, bool, str, date, datetime, PyLegendPrimitive]],
            ]
    ) -> None:
        self.__base_frame = base_frame
        self.__col_definitions = col_definitions

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        db_extension = config.sql_to_string_generator().get_db_extension()
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (len(base_query.groupBy) > 0) or base_query.select.distinct

        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )

        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        for col, func in self.__col_definitions.items():
            res = func(tds_row)
            if not isinstance(res, PyLegendPrimitive):
                raise RuntimeError("Constants not supported")
            new_col_expr = res.to_sql_expression(
                {"frame": base_query},
                config
            )
            new_query.select.selectItems.append(
                SingleColumn(alias=db_extension.quote_identifier(col), expression=new_col_expr)
            )
        return new_query

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        new_cols = [c.copy() for c in self.__base_frame.columns()]
        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        for col, func in self.__col_definitions.items():
            res = func(tds_row)
            if isinstance(res, (int, PyLegendInteger)):
                new_cols.append(PrimitiveTdsColumn.integer_column(col))
            elif isinstance(res, (float, PyLegendFloat)):
                new_cols.append(PrimitiveTdsColumn.float_column(col))
            elif isinstance(res, PyLegendNumber):
                new_cols.append(PrimitiveTdsColumn.number_column(col))
            elif isinstance(res, (bool, PyLegendBoolean)):
                new_cols.append(PrimitiveTdsColumn.boolean_column(col))
            elif isinstance(res, (str, PyLegendString)):
                new_cols.append(PrimitiveTdsColumn.string_column(col))
            else:
                raise RuntimeError("Type not supported")
        return new_cols

    def validate(self) -> bool:
        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        for col, f in self.__col_definitions.items():
            f(tds_row)
        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
)
from pylegend.core.tds.tds_frame import (
    PyLegendTdsFrame,
    FrameToSqlConfig,
)
from pylegend.core.project_cooridnates import ProjectCoordinates
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    TableFunction,
    Select,
    AllColumns,
    FunctionCall,
    QualifiedName,
    NamedArgumentExpression,
    StringLiteral,
    AliasedRelation,
    SingleColumn,
    QualifiedNameReference,
    Expression,
)

__all__: PyLegendSequence[str] = [
    "LegendServiceInputFrameAbstract",
]


class LegendServiceInputFrameAbstract(PyLegendTdsFrame, metaclass=ABCMeta):
    __pattern: str
    __project_coordinates: ProjectCoordinates
    __initialized: bool = False

    def __init__(
            self,
            pattern: str,
            project_coordinates: ProjectCoordinates,
    ) -> None:
        self.__pattern = pattern
        self.__project_coordinates = project_coordinates

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        db_extension = config.sql_to_string_generator().get_db_extension()
        root_alias = db_extension.quote_identifier("root")
        args: PyLegendList[Expression] = [
            NamedArgumentExpression(
                name="pattern",
                expression=StringLiteral(value=self.__pattern, quoted=False)
            )
        ]
        args += self.__project_coordinates.sql_params()
        func_call = FunctionCall(
            name=QualifiedName(["service"]),
            distinct=False,
            filter_=None,
            window=None,
            arguments=args
        )

        return QuerySpecification(
            select=Select(
                selectItems=[
                    SingleColumn(
                        alias=db_extension.quote_identifier(x.get_name()),
                        expression=QualifiedNameReference(
                            name=QualifiedName(parts=[root_alias, db_extension.quote_identifier(x.get_name())])
                        )
                    )
                    for x in self.columns()
                ] if self.__initialized else [AllColumns(prefix=root_alias)],
                distinct=False
            ),
            from_=[
                AliasedRelation(
                    relation=TableFunction(functionCall=func_call),
                    alias=root_alias,
                    columnNames=[x.get_name() for x in self.columns()] if self.__initialized else []
                )
            ],
            where=None,
            groupBy=[],
            having=None,
            orderBy=[],
            limit=None,
            offset=None
        )

    def get_pattern(self) -> str:
        return self.__pattern

    def set_initialized(self, val: bool) -> None:
        self.__initialized = val



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.expression import (
    PyLegendExpressionNumberReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionBooleanReturn,
)
from pylegend.core.language.operations.binary_expression import PyLegendBinaryExpression
from pylegend.core.language.operations.unary_expression import PyLegendUnaryExpression
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
    ArithmeticType,
    ArithmeticExpression,
    ComparisonOperator,
    ComparisonExpression,
    NegativeExpression,
)
from pylegend.core.sql.metamodel_extension import (
    AbsoluteExpression,
    PowerExpression,
    CeilExpression,
    FloorExpression,
    SqrtExpression,
    CbrtExpression,
    ExpExpression,
    LogExpression,
    RemainderExpression,
    RoundExpression,
    SineExpression,
    ArcSineExpression,
    CosineExpression,
    ArcCosineExpression,
    TanExpression,
    ArcTanExpression,
    ArcTan2Expression,
    CotExpression,
)


__all__: PyLegendSequence[str] = [
    "PyLegendNumberAddExpression",
    "PyLegendNumberMultiplyExpression",
    "PyLegendNumberDivideExpression",
    "PyLegendNumberSubtractExpression",
    "PyLegendNumberLessThanExpression",
    "PyLegendNumberLessThanEqualExpression",
    "PyLegendNumberGreaterThanExpression",
    "PyLegendNumberGreaterThanEqualExpression",
    "PyLegendNumberNegativeExpression",
    "PyLegendNumberAbsoluteExpression",
    "PyLegendNumberPowerExpression",
    "PyLegendNumberCeilExpression",
    "PyLegendNumberFloorExpression",
    "PyLegendNumberSqrtExpression",
    "PyLegendNumberCbrtExpression",
    "PyLegendNumberExpExpression",
    "PyLegendNumberLogExpression",
    "PyLegendNumberRemainderExpression",
    "PyLegendNumberRoundExpression",
    "PyLegendNumberSineExpression",
    "PyLegendNumberArcSineExpression",
    "PyLegendNumberCosineExpression",
    "PyLegendNumberArcCosineExpression",
    "PyLegendNumberTanExpression",
    "PyLegendNumberArcTanExpression",
    "PyLegendNumberArcTan2Expression",
    "PyLegendNumberCotExpression",
]


class PyLegendNumberAddExpression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.ADD, expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberAddExpression.__to_sql_func
        )


class PyLegendNumberMultiplyExpression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.MULTIPLY, expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberMultiplyExpression.__to_sql_func
        )


class PyLegendNumberDivideExpression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.DIVIDE, expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberDivideExpression.__to_sql_func
        )


class PyLegendNumberSubtractExpression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArithmeticExpression(ArithmeticType.SUBTRACT, expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberSubtractExpression.__to_sql_func
        )


class PyLegendNumberLessThanExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.LESS_THAN)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberLessThanExpression.__to_sql_func
        )


class PyLegendNumberLessThanEqualExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.LESS_THAN_OR_EQUAL)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberLessThanEqualExpression.__to_sql_func
        )


class PyLegendNumberGreaterThanExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.GREATER_THAN)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberGreaterThanExpression.__to_sql_func
        )


class PyLegendNumberGreaterThanEqualExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.GREATER_THAN_OR_EQUAL)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberGreaterThanEqualExpression.__to_sql_func
        )


class PyLegendNumberNegativeExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return NegativeExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberNegativeExpression.__to_sql_func
        )


class PyLegendNumberAbsoluteExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return AbsoluteExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberAbsoluteExpression.__to_sql_func
        )


class PyLegendNumberPowerExpression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return PowerExpression(expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberPowerExpression.__to_sql_func
        )


class PyLegendNumberCeilExpression(PyLegendUnaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return CeilExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberCeilExpression.__to_sql_func
        )


class PyLegendNumberFloorExpression(PyLegendUnaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FloorExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberFloorExpression.__to_sql_func
        )


class PyLegendNumberSqrtExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return SqrtExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberSqrtExpression.__to_sql_func
        )


class PyLegendNumberCbrtExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return CbrtExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberCbrtExpression.__to_sql_func
        )


class PyLegendNumberExpExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ExpExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberExpExpression.__to_sql_func
        )


class PyLegendNumberLogExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return LogExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberLogExpression.__to_sql_func
        )


class PyLegendNumberRemainderExpression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return RemainderExpression(expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberRemainderExpression.__to_sql_func
        )


class PyLegendNumberRoundExpression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return RoundExpression(expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionIntegerReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberRoundExpression.__to_sql_func
        )


class PyLegendNumberSineExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return SineExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberSineExpression.__to_sql_func
        )


class PyLegendNumberArcSineExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArcSineExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberArcSineExpression.__to_sql_func
        )


class PyLegendNumberCosineExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return CosineExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberCosineExpression.__to_sql_func
        )


class PyLegendNumberArcCosineExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArcCosineExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberArcCosineExpression.__to_sql_func
        )


class PyLegendNumberTanExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return TanExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberTanExpression.__to_sql_func
        )


class PyLegendNumberArcTanExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArcTanExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberArcTanExpression.__to_sql_func
        )


class PyLegendNumberArcTan2Expression(PyLegendBinaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ArcTan2Expression(expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionNumberReturn, operand2: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendNumberArcTan2Expression.__to_sql_func
        )


class PyLegendNumberCotExpression(PyLegendUnaryExpression, PyLegendExpressionNumberReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return CotExpression(expression)

    def __init__(self, operand: PyLegendExpressionNumberReturn) -> None:
        PyLegendExpressionNumberReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendNumberCotExpression.__to_sql_func
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
)
from pylegend.core.language.primitives.primitive import PyLegendPrimitive
from pylegend.core.language.primitives.integer import PyLegendInteger
from pylegend.core.language.primitives.float import PyLegendFloat
from pylegend.core.language.primitives.boolean import PyLegendBoolean
from pylegend.core.language.expression import PyLegendExpressionStringReturn, PyLegendExpression
from pylegend.core.language.literal_expressions import PyLegendStringLiteralExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.language.operations.string_operation_expressions import (
    PyLegendStringLengthExpression,
    PyLegendStringLikeExpression,
    PyLegendStringUpperExpression,
    PyLegendStringLowerExpression,
    PyLegendStringLTrimExpression,
    PyLegendStringRTrimExpression,
    PyLegendStringBTrimExpression,
    PyLegendStringPosExpression,
    PyLegendStringParseIntExpression,
    PyLegendStringParseFloatExpression,
    PyLegendStringConcatExpression,
    PyLegendStringLessThanExpression,
    PyLegendStringLessThanEqualExpression,
    PyLegendStringGreaterThanExpression,
    PyLegendStringGreaterThanEqualExpression,
)


__all__: PyLegendSequence[str] = [
    "PyLegendString"
]


class PyLegendString(PyLegendPrimitive):
    __value: PyLegendExpressionStringReturn

    def __init__(
            self,
            value: PyLegendExpressionStringReturn
    ) -> None:
        self.__value = value

    def len(self) -> PyLegendInteger:
        return PyLegendInteger(PyLegendStringLengthExpression(self.__value))

    def length(self) -> PyLegendInteger:
        return self.len()

    def startswith(self, prefix: str) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str(prefix, "startswith prefix parameter")
        escaped_prefix = PyLegendString.__escape_like_param(prefix)
        return PyLegendBoolean(
            PyLegendStringLikeExpression(self.__value, PyLegendStringLiteralExpression(escaped_prefix + '%'))
        )

    def endswith(self, suffix: str) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str(suffix, "endswith suffix parameter")
        escaped_suffix = PyLegendString.__escape_like_param(suffix)
        return PyLegendBoolean(
            PyLegendStringLikeExpression(self.__value, PyLegendStringLiteralExpression("%" + escaped_suffix))
        )

    def contains(self, other: str) -> PyLegendBoolean:
        PyLegendString.__validate_param_to_be_str(other, "contains/in other parameter")
        escaped_other = PyLegendString.__escape_like_param(other)
        return PyLegendBoolean(
            PyLegendStringLikeExpression(self.__value, PyLegendStringLiteralExpression("%" + escaped_other + "%"))
        )

    def upper(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringUpperExpression(self.__value))

    def lower(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringLowerExpression(self.__value))

    def lstrip(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringLTrimExpression(self.__value))

    def rstrip(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringRTrimExpression(self.__value))

    def strip(self) -> "PyLegendString":
        return PyLegendString(PyLegendStringBTrimExpression(self.__value))

    def index_of(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendInteger":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "Index_of parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendInteger(PyLegendStringPosExpression(self.__value, other_op))

    def index(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendInteger":
        return self.index_of(other)

    def parse_int(self) -> "PyLegendInteger":
        return PyLegendInteger(PyLegendStringParseIntExpression(self.__value))

    def parse_integer(self) -> "PyLegendInteger":
        return self.parse_int()

    def parse_float(self) -> "PyLegendFloat":
        return PyLegendFloat(PyLegendStringParseFloatExpression(self.__value))

    def __add__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String plus (+) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendString(PyLegendStringConcatExpression(self.__value, other_op))

    def __radd__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendString":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String plus (+) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendString(PyLegendStringConcatExpression(other_op, self.__value))

    def __lt__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String less than (<) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringLessThanExpression(self.__value, other_op))

    def __le__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String less than equal (<=) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringLessThanEqualExpression(self.__value, other_op))

    def __gt__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String greater than (>) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringGreaterThanExpression(self.__value, other_op))

    def __ge__(self, other: PyLegendUnion[str, "PyLegendString"]) -> "PyLegendBoolean":
        PyLegendString.__validate_param_to_be_str_or_str_expr(other, "String greater than equal (>=) parameter")
        other_op = PyLegendStringLiteralExpression(other) if isinstance(other, str) else other.__value
        return PyLegendBoolean(PyLegendStringGreaterThanEqualExpression(self.__value, other_op))

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return self.__value.to_sql_expression(frame_name_to_base_query_map, config)

    def value(self) -> PyLegendExpression:
        return self.__value

    @staticmethod
    def __escape_like_param(param: str) -> str:
        return param.replace("_", "\\_").replace("%", "\\%")

    @staticmethod
    def __validate_param_to_be_str_or_str_expr(param: PyLegendUnion[str, "PyLegendString"], desc: str) -> None:
        if not isinstance(param, (str, PyLegendString)):
            raise TypeError(desc + " should be a str or a string expression (PyLegendString)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))

    @staticmethod
    def __validate_param_to_be_str(param: str, desc: str) -> None:
        if not isinstance(param, str):
            raise TypeError(desc + " should be a str."
                                   " Got value " + str(param) + " of type: " + str(type(param)))



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence
)
from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import LegendApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    LongLiteral,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_base_tds_frame import LegendApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "SliceFunction"
]


class SliceFunction(LegendApiAppliedFunction):
    __base_frame: LegendApiBaseTdsFrame
    __start_row: int
    __end_row: int

    @classmethod
    def name(cls) -> str:
        return "slice"

    def __init__(self, base_frame: LegendApiBaseTdsFrame, start_row: int, end_row: int) -> None:
        self.__base_frame = base_frame
        self.__start_row = start_row
        self.__end_row = end_row

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (base_query.offset is not None) or (base_query.limit is not None)
        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query else
            copy_query(base_query)
        )
        new_query.offset = LongLiteral(self.__start_row)
        new_query.limit = LongLiteral(self.__end_row - self.__start_row)
        return new_query

    def base_frame(self) -> LegendApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        if self.__start_row < 0:
            raise ValueError(
                "Start row argument of slice function cannot be negative. Start row: " + str(self.__start_row)
            )
        if self.__end_row <= self.__start_row:
            raise ValueError("End row argument of slice function cannot be less than or equal to start row argument. "
                             f"Start row: {self.__start_row}, End row: {self.__end_row}")
        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
import pandas as pd
from pylegend._typing import (
    PyLegendSequence,
    PyLegendTypeVar,
    PyLegendList,
    PyLegendOptional,
    PyLegendCallable,
    PyLegendUnion,
)
from pylegend.core.sql.metamodel import QuerySpecification
from pylegend.core.databse.sql_to_string import (
    SqlToStringConfig,
    SqlToStringFormat
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_tds_frame import LegendApiTdsFrame
from pylegend.core.tds.result_handler import (
    ResultHandler,
    ToStringResultHandler,
)
from pylegend.extensions.tds.result_handler import (
    ToPandasDfResultHandler,
    PandasDfReadConfig,
)
from pylegend.core.language import (
    TdsRow,
    PyLegendBoolean,
    PyLegendPrimitiveOrPythonPrimitive,
    AggregateSpecification,
)

__all__: PyLegendSequence[str] = [
    "LegendApiBaseTdsFrame"
]

R = PyLegendTypeVar('R')


class LegendApiBaseTdsFrame(LegendApiTdsFrame, metaclass=ABCMeta):
    __columns: PyLegendSequence[TdsColumn]

    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        col_names = [c.get_name() for c in columns]
        if len(col_names) != len(set(col_names)):
            cols = "[" + ", ".join([str(c) for c in columns]) + "]"
            raise ValueError(f"TdsFrame cannot have duplicated column names. Passed columns: {cols}")
        self.__columns = [c.copy() for c in columns]

    def columns(self) -> PyLegendSequence[TdsColumn]:
        return [c.copy() for c in self.__columns]

    def head(self, row_count: int = 5) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.head_function import (
            HeadFunction
        )
        return LegendApiAppliedFunctionTdsFrame(HeadFunction(self, row_count))

    def take(self, row_count: int = 5) -> "LegendApiTdsFrame":
        return self.head(row_count=row_count)

    def limit(self, row_count: int = 5) -> "LegendApiTdsFrame":
        return self.head(row_count=row_count)

    def drop(self, row_count: int = 5) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.drop_function import (
            DropFunction
        )
        return LegendApiAppliedFunctionTdsFrame(DropFunction(self, row_count))

    def slice(self, start_row: int, end_row_exclusive: int) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.slice_function import (
            SliceFunction
        )
        return LegendApiAppliedFunctionTdsFrame(SliceFunction(self, start_row, end_row_exclusive))

    def distinct(self) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.distinct_function import (
            DistinctFunction
        )
        return LegendApiAppliedFunctionTdsFrame(DistinctFunction(self))

    def restrict(self, column_name_list: PyLegendList[str]) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.restrict_function import (
            RestrictFunction
        )
        return LegendApiAppliedFunctionTdsFrame(RestrictFunction(self, column_name_list))

    def sort(
            self,
            column_name_list: PyLegendList[str],
            direction_list: PyLegendOptional[PyLegendList[str]] = None
    ) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.sort_function import (
            SortFunction
        )
        return LegendApiAppliedFunctionTdsFrame(SortFunction(self, column_name_list, direction_list))

    def concatenate(self, other: "LegendApiTdsFrame") -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.concatenate_function import (
            ConcatenateFunction
        )
        return LegendApiAppliedFunctionTdsFrame(ConcatenateFunction(self, other))

    def rename_columns(
            self,
            column_names: PyLegendList[str],
            renamed_column_names: PyLegendList[str]
    ) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.rename_columns_function import (
            RenameColumnsFunction
        )
        return LegendApiAppliedFunctionTdsFrame(RenameColumnsFunction(self, column_names, renamed_column_names))

    def filter(
            self,
            filter_function: PyLegendCallable[[TdsRow], PyLegendUnion[bool, PyLegendBoolean]]
    ) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.filter_function import (
            FilterFunction
        )
        return LegendApiAppliedFunctionTdsFrame(FilterFunction(self, filter_function))

    def extend(
            self,
            functions_list: PyLegendList[PyLegendCallable[[TdsRow], PyLegendPrimitiveOrPythonPrimitive]],
            column_names_list: PyLegendList[str]
    ) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.extend_function import (
            ExtendFunction
        )
        return LegendApiAppliedFunctionTdsFrame(ExtendFunction(self, functions_list, column_names_list))

    def join(
            self,
            other: "LegendApiTdsFrame",
            join_condition: PyLegendCallable[[TdsRow, TdsRow], PyLegendUnion[bool, PyLegendBoolean]],
            join_type: str = 'LEFT_OUTER'
    ) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.join_function import (
            JoinFunction
        )
        return LegendApiAppliedFunctionTdsFrame(
            JoinFunction(self, other, join_condition, join_type)
        )

    def join_by_columns(
            self,
            other: "LegendApiTdsFrame",
            self_columns: PyLegendList[str],
            other_columns: PyLegendList[str],
            join_type: str = 'LEFT_OUTER'
    ) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.join_by_columns_function import (
            JoinByColumnsFunction
        )
        return LegendApiAppliedFunctionTdsFrame(
            JoinByColumnsFunction(self, other, self_columns, other_columns, join_type)
        )

    def group_by(
            self,
            grouping_columns: PyLegendList[str],
            aggregations: PyLegendList[AggregateSpecification],
    ) -> "LegendApiTdsFrame":
        from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import (
            LegendApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.legend_api.frames.functions.group_by_function import (
            GroupByFunction
        )
        return LegendApiAppliedFunctionTdsFrame(
            GroupByFunction(self, grouping_columns, aggregations)
        )

    @abstractmethod
    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        pass  # pragma: no cover

    @abstractmethod
    def get_all_tds_frames(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        pass  # pragma: no cover

    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        query = self.to_sql_query_object(config)
        sql_to_string_config = SqlToStringConfig(
            format_=SqlToStringFormat(
                pretty=config.pretty
            )
        )
        return config.sql_to_string_generator().generate_sql_string(query, sql_to_string_config)

    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:
        from pylegend.core.tds.legend_api.frames.legend_api_input_tds_frame import (
            LegendApiInputTdsFrame,
            LegendApiExecutableInputTdsFrame
        )

        tds_frames = self.get_all_tds_frames()
        input_frames = [x for x in tds_frames if isinstance(x, LegendApiInputTdsFrame)]

        non_exec_frames = [x for x in input_frames if not isinstance(x, LegendApiExecutableInputTdsFrame)]
        if non_exec_frames:
            raise ValueError(
                "Cannot execute frame as its built on top of non-executable input frames: [" +
                (", ".join([str(f) for f in non_exec_frames]) + "]")
            )

        exec_frames = [x for x in input_frames if isinstance(x, LegendApiExecutableInputTdsFrame)]

        all_legend_clients = []
        for e in exec_frames:
            c = e.get_legend_client()
            if c not in all_legend_clients:
                all_legend_clients.append(c)
        if len(all_legend_clients) > 1:
            raise ValueError(
                "Found tds frames with multiple legend_clients (which is not supported): [" +
                (", ".join([str(f) for f in all_legend_clients]) + "]")
            )
        legend_client = all_legend_clients[0]
        result = legend_client.execute_sql_string(self.to_sql_query(), chunk_size=chunk_size)
        return result_handler.handle_result(self, result)

    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:
        return self.execute_frame(ToStringResultHandler(), chunk_size)

    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:
        return self.execute_frame(
            ToPandasDfResultHandler(pandas_df_read_config),
            chunk_size
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.tds.tds_frame import PyLegendTdsFrame
from pylegend.core.tds.tds_column import TdsColumn, PrimitiveTdsColumn
from pylegend.core.language import (
    PyLegendColumnExpression,
    PyLegendBooleanColumnExpression,
    PyLegendStringColumnExpression,
    PyLegendNumberColumnExpression,
    PyLegendIntegerColumnExpression,
    PyLegendFloatColumnExpression,
    PyLegendDateColumnExpression,
    PyLegendDateTimeColumnExpression,
    PyLegendStrictDateColumnExpression,
    PyLegendPrimitive,
    PyLegendBoolean,
    PyLegendString,
    PyLegendNumber,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendDate,
    PyLegendDateTime,
    PyLegendStrictDate,
)

__all__: PyLegendSequence[str] = [
    "TdsRow",
]


class TdsRow:
    __frame_name: str
    __columns: PyLegendSequence[TdsColumn]

    def __init__(self, frame_name: str, frame: PyLegendTdsFrame) -> None:
        self.__frame_name = frame_name
        self.__columns = frame.columns()

    @staticmethod
    def from_tds_frame(frame_name: str, frame: PyLegendTdsFrame) -> "TdsRow":
        return TdsRow(frame_name=frame_name, frame=frame)

    def get_boolean(self, column: str) -> PyLegendBoolean:
        col_expr = self.__get_col(column)
        if not isinstance(col_expr, PyLegendBooleanColumnExpression):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_boolean method is not valid on this column."
            )
        return PyLegendBoolean(col_expr)

    def get_string(self, column: str) -> PyLegendString:
        col_expr = self.__get_col(column)
        if not isinstance(col_expr, PyLegendStringColumnExpression):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_string method is not valid on this column."
            )
        return PyLegendString(col_expr)

    def get_number(self, column: str) -> PyLegendNumber:
        col_expr = self.__get_col(column)
        allowed_types = (PyLegendNumberColumnExpression, PyLegendIntegerColumnExpression, PyLegendFloatColumnExpression)
        if not isinstance(col_expr, allowed_types):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_number method is not valid on this column."
            )
        return PyLegendNumber(col_expr)

    def get_integer(self, column: str) -> PyLegendInteger:
        col_expr = self.__get_col(column)
        if not isinstance(col_expr, PyLegendIntegerColumnExpression):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_integer method is not valid on this column."
            )
        return PyLegendInteger(col_expr)

    def get_float(self, column: str) -> PyLegendFloat:
        col_expr = self.__get_col(column)
        if not isinstance(col_expr, PyLegendFloatColumnExpression):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_float method is not valid on this column."
            )
        return PyLegendFloat(col_expr)

    def get_date(self, column: str) -> PyLegendDate:
        col_expr = self.__get_col(column)
        allowed_types = (
            PyLegendDateColumnExpression, PyLegendDateTimeColumnExpression, PyLegendStrictDateColumnExpression
        )
        if not isinstance(col_expr, allowed_types):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_date method is not valid on this column."
            )
        return PyLegendDate(col_expr)

    def get_datetime(self, column: str) -> PyLegendDateTime:
        col_expr = self.__get_col(column)
        if not isinstance(col_expr, PyLegendDateTimeColumnExpression):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_datetime method is not valid on this column."
            )
        return PyLegendDateTime(col_expr)

    def get_strictdate(self, column: str) -> PyLegendStrictDate:
        col_expr = self.__get_col(column)
        if not isinstance(col_expr, PyLegendStrictDateColumnExpression):
            raise RuntimeError(
                f"Column expression for '{column}' is of type '{type(col_expr)}'. "
                "get_strictdate method is not valid on this column."
            )
        return PyLegendStrictDate(col_expr)

    def __getitem__(self, item: str) -> PyLegendPrimitive:
        if not isinstance(item, str):
            raise TypeError("Column indexing on a TDSRow should be with column name (string). Got - " + str(type(item)))

        col_expr = self.__get_col(item)

        if isinstance(col_expr, PyLegendBooleanColumnExpression):
            return PyLegendBoolean(col_expr)
        if isinstance(col_expr, PyLegendStringColumnExpression):
            return PyLegendString(col_expr)
        if isinstance(col_expr, PyLegendIntegerColumnExpression):
            return PyLegendInteger(col_expr)
        if isinstance(col_expr, PyLegendFloatColumnExpression):
            return PyLegendFloat(col_expr)
        if isinstance(col_expr, PyLegendNumberColumnExpression):
            return PyLegendNumber(col_expr)
        if isinstance(col_expr, PyLegendDateTimeColumnExpression):
            return PyLegendDateTime(col_expr)
        if isinstance(col_expr, PyLegendStrictDateColumnExpression):
            return PyLegendStrictDate(col_expr)
        if isinstance(col_expr, PyLegendDateColumnExpression):
            return PyLegendDate(col_expr)

        raise RuntimeError(f"Column expression for '{item}' of type {type(col_expr)} not supported yet")

    def __get_col(self, column: str) -> PyLegendColumnExpression:
        for base_col in self.__columns:
            if base_col.get_name() == column:
                if isinstance(base_col, PrimitiveTdsColumn):
                    if base_col.get_type() == "Boolean":
                        return PyLegendBooleanColumnExpression(self.__frame_name, column)
                    if base_col.get_type() == "String":
                        return PyLegendStringColumnExpression(self.__frame_name, column)
                    if base_col.get_type() == "Number":
                        return PyLegendNumberColumnExpression(self.__frame_name, column)
                    if base_col.get_type() == "Integer":
                        return PyLegendIntegerColumnExpression(self.__frame_name, column)
                    if base_col.get_type() == "Float":
                        return PyLegendFloatColumnExpression(self.__frame_name, column)
                    if base_col.get_type() == "Date":
                        return PyLegendDateColumnExpression(self.__frame_name, column)
                    if base_col.get_type() == "DateTime":
                        return PyLegendDateTimeColumnExpression(self.__frame_name, column)
                    if base_col.get_type() == "StrictDate":
                        return PyLegendStrictDateColumnExpression(self.__frame_name, column)

                raise RuntimeError(f"Column '{column}' of type {base_col.get_type()} not supported yet")

        raise ValueError(
            f"Column - '{column}' doesn't exist in the current frame. "
            f"Current frame columns: {[x.get_name() for x in self.__columns]}"
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.expression import (
    PyLegendExpressionStringReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionBooleanReturn,
)
from pylegend.core.language.operations.binary_expression import PyLegendBinaryExpression
from pylegend.core.language.operations.unary_expression import PyLegendUnaryExpression
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
    ColumnType,
    Cast,
    ComparisonOperator,
    ComparisonExpression,
)
from pylegend.core.sql.metamodel_extension import (
    StringLengthExpression,
    StringLikeExpression,
    StringUpperExpression,
    StringLowerExpression,
    TrimType,
    StringTrimExpression,
    StringPosExpression,
    StringConcatExpression,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig


__all__: PyLegendSequence[str] = [
    "PyLegendStringLengthExpression",
    "PyLegendStringLikeExpression",
    "PyLegendStringUpperExpression",
    "PyLegendStringLowerExpression",
    "PyLegendStringLTrimExpression",
    "PyLegendStringRTrimExpression",
    "PyLegendStringBTrimExpression",
    "PyLegendStringPosExpression",
    "PyLegendStringParseIntExpression",
    "PyLegendStringParseFloatExpression",
    "PyLegendStringConcatExpression",
    "PyLegendStringLessThanExpression",
    "PyLegendStringLessThanEqualExpression",
    "PyLegendStringGreaterThanExpression",
    "PyLegendStringGreaterThanEqualExpression",
]


class PyLegendStringLengthExpression(PyLegendUnaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringLengthExpression(expression)

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringLengthExpression.__to_sql_func
        )


class PyLegendStringLikeExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringLikeExpression(expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionStringReturn, operand2: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendStringLikeExpression.__to_sql_func
        )


class PyLegendStringUpperExpression(PyLegendUnaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringUpperExpression(expression)

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringUpperExpression.__to_sql_func
        )


class PyLegendStringLowerExpression(PyLegendUnaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringLowerExpression(expression)

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringLowerExpression.__to_sql_func
        )


class PyLegendStringLTrimExpression(PyLegendUnaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringTrimExpression(expression, trim_type=TrimType.Left)

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringLTrimExpression.__to_sql_func
        )


class PyLegendStringRTrimExpression(PyLegendUnaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringTrimExpression(expression, trim_type=TrimType.Right)

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringRTrimExpression.__to_sql_func
        )


class PyLegendStringBTrimExpression(PyLegendUnaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringTrimExpression(expression, trim_type=TrimType.Both)

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringBTrimExpression.__to_sql_func
        )


class PyLegendStringPosExpression(PyLegendBinaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringPosExpression(expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionStringReturn, operand2: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendStringPosExpression.__to_sql_func
        )


class PyLegendStringParseIntExpression(PyLegendUnaryExpression, PyLegendExpressionIntegerReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return Cast(expression, ColumnType(name="INTEGER", parameters=[]))

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionIntegerReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringParseIntExpression.__to_sql_func
        )


class PyLegendStringParseFloatExpression(PyLegendUnaryExpression, PyLegendExpressionFloatReturn):

    @staticmethod
    def __to_sql_func(
            expression: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return Cast(expression, ColumnType(name="DOUBLE PRECISION", parameters=[]))

    def __init__(self, operand: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionFloatReturn.__init__(self)
        PyLegendUnaryExpression.__init__(
            self,
            operand,
            PyLegendStringParseFloatExpression.__to_sql_func
        )


class PyLegendStringConcatExpression(PyLegendBinaryExpression, PyLegendExpressionStringReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return StringConcatExpression(expression1, expression2)

    def __init__(self, operand1: PyLegendExpressionStringReturn, operand2: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionStringReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendStringConcatExpression.__to_sql_func
        )


class PyLegendStringLessThanExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.LESS_THAN)

    def __init__(self, operand1: PyLegendExpressionStringReturn, operand2: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendStringLessThanExpression.__to_sql_func
        )


class PyLegendStringLessThanEqualExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.LESS_THAN_OR_EQUAL)

    def __init__(self, operand1: PyLegendExpressionStringReturn, operand2: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendStringLessThanEqualExpression.__to_sql_func
        )


class PyLegendStringGreaterThanExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.GREATER_THAN)

    def __init__(self, operand1: PyLegendExpressionStringReturn, operand2: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendStringGreaterThanExpression.__to_sql_func
        )


class PyLegendStringGreaterThanEqualExpression(PyLegendBinaryExpression, PyLegendExpressionBooleanReturn):

    @staticmethod
    def __to_sql_func(
            expression1: Expression,
            expression2: Expression,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return ComparisonExpression(expression1, expression2, ComparisonOperator.GREATER_THAN_OR_EQUAL)

    def __init__(self, operand1: PyLegendExpressionStringReturn, operand2: PyLegendExpressionStringReturn) -> None:
        PyLegendExpressionBooleanReturn.__init__(self)
        PyLegendBinaryExpression.__init__(
            self,
            operand1,
            operand2,
            PyLegendStringGreaterThanEqualExpression.__to_sql_func
        )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
)
from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import LegendApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query, extract_columns_for_subquery
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    Select,
    SelectItem,
    SingleColumn,
    AliasedRelation,
    TableSubquery,
    Query,
    Join,
    JoinType,
    JoinOn,
    QualifiedNameReference,
    QualifiedName,
    LogicalBinaryExpression,
    LogicalBinaryType,
    ComparisonExpression,
    ComparisonOperator,
    Expression,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_base_tds_frame import LegendApiBaseTdsFrame
from pylegend.core.tds.legend_api.frames.legend_api_tds_frame import LegendApiTdsFrame


__all__: PyLegendSequence[str] = [
    "JoinByColumnsFunction"
]


class JoinByColumnsFunction(LegendApiAppliedFunction):
    __base_frame: LegendApiBaseTdsFrame
    __other_frame: LegendApiBaseTdsFrame
    __column_names_self: PyLegendList[str]
    __column_names_other: PyLegendList[str]
    __join_type: str

    @classmethod
    def name(cls) -> str:
        return "join_by_columns"

    def __init__(
            self,
            base_frame: LegendApiBaseTdsFrame,
            other_frame: LegendApiTdsFrame,
            column_names_self: PyLegendList[str],
            column_names_other: PyLegendList[str],
            join_type: str
    ) -> None:
        self.__base_frame = base_frame
        if not isinstance(other_frame, LegendApiBaseTdsFrame):
            raise ValueError("Expected LegendApiBaseTdsFrame")  # pragma: no cover
        self.__other_frame = other_frame
        self.__column_names_self = column_names_self
        self.__column_names_other = column_names_other
        self.__join_type = join_type

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        db_extension = config.sql_to_string_generator().get_db_extension()
        base_query = copy_query(self.__base_frame.to_sql_query_object(config))
        other_query = copy_query(self.__other_frame.to_sql_query_object(config))
        left_alias = db_extension.quote_identifier('left')
        right_alias = db_extension.quote_identifier('right')

        join_type = (
            JoinType.INNER if self.__join_type.lower() == 'inner' else (
                JoinType.LEFT if self.__join_type.lower() in ('left_outer', 'leftouter') else
                JoinType.RIGHT
            )
        )

        def logical_and_expr(left: Expression, right: Expression) -> Expression:
            return LogicalBinaryExpression(
                type_=LogicalBinaryType.AND,
                left=left,
                right=right
            )
        join_expr = functools.reduce(
            logical_and_expr,  # type: ignore
            [
                ComparisonExpression(
                    left=QualifiedNameReference(
                        name=QualifiedName(parts=[left_alias, db_extension.quote_identifier(x)])
                    ),
                    right=QualifiedNameReference(
                        name=QualifiedName(parts=[right_alias, db_extension.quote_identifier(y)])
                    ),
                    operator=ComparisonOperator.EQUAL
                )
                for x, y in zip(self.__column_names_self, self.__column_names_other)
            ]
        )

        common_join_cols = [x for x, y in zip(self.__column_names_self, self.__column_names_other) if x == y]
        common_join_cols.sort()
        new_select_items: PyLegendList[SelectItem] = []
        for c in (c for c in self.__base_frame.columns() if c.get_name() not in common_join_cols):
            q = db_extension.quote_identifier(c.get_name())
            new_select_items.append(SingleColumn(q, QualifiedNameReference(name=QualifiedName(parts=[left_alias, q]))))
        for c in (c for c in self.__base_frame.columns() if c.get_name() in common_join_cols):
            q = db_extension.quote_identifier(c.get_name())
            common_col_alias = right_alias if self.__join_type.lower() == 'right_outer' else left_alias
            new_select_items.append(
                SingleColumn(q, QualifiedNameReference(name=QualifiedName(parts=[common_col_alias, q])))
            )
        for c in (c for c in self.__other_frame.columns() if c.get_name() not in common_join_cols):
            q = db_extension.quote_identifier(c.get_name())
            new_select_items.append(SingleColumn(q, QualifiedNameReference(name=QualifiedName(parts=[right_alias, q]))))

        join_query = QuerySpecification(
            select=Select(
                selectItems=new_select_items,
                distinct=False
            ),
            from_=[
                Join(
                    type_=join_type,
                    left=AliasedRelation(
                        relation=TableSubquery(query=Query(queryBody=base_query, limit=None, offset=None, orderBy=[])),
                        alias=left_alias,
                        columnNames=extract_columns_for_subquery(base_query)
                    ),
                    right=AliasedRelation(
                        relation=TableSubquery(query=Query(queryBody=other_query, limit=None, offset=None, orderBy=[])),
                        alias=right_alias,
                        columnNames=extract_columns_for_subquery(other_query)
                    ),
                    criteria=JoinOn(expression=join_expr)
                )
            ],
            where=None,
            groupBy=[],
            having=None,
            orderBy=[],
            limit=None,
            offset=None
        )

        wrapped_join_query = create_sub_query(join_query, config, "root")
        return wrapped_join_query

    def base_frame(self) -> LegendApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        return [self.__other_frame]

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        common_join_cols = [x for x, y in zip(self.__column_names_self, self.__column_names_other) if x == y]
        common_join_cols.sort()
        return (
            [c.copy() for c in self.__base_frame.columns() if c.get_name() not in common_join_cols] +
            [c.copy() for c in self.__base_frame.columns() if c.get_name() in common_join_cols] +
            [c.copy() for c in self.__other_frame.columns() if c.get_name() not in common_join_cols]
        )

    def validate(self) -> bool:
        left_cols = [c.get_name() for c in self.__base_frame.columns()]
        for c in self.__column_names_self:
            if c not in left_cols:
                raise ValueError(
                    f"Column - '{c}' in join columns list doesn't exist in the left frame being joined. "
                    f"Current left frame columns: {left_cols}"
                )

        right_cols = [c.get_name() for c in self.__other_frame.columns()]
        for c in self.__column_names_other:
            if c not in right_cols:
                raise ValueError(
                    f"Column - '{c}' in join columns list doesn't exist in the right frame being joined. "
                    f"Current right frame columns: {right_cols}"
                )

        if len(self.__column_names_self) != len(self.__column_names_other):
            raise ValueError(
                "For join_by_columns function, column lists should be of same size. "
                f"Passed column list sizes -  Left: {len(self.__column_names_self)}, Right: {len(self.__column_names_other)}"
            )

        if len(self.__column_names_self) == 0:
            raise ValueError("For join_by_columns function, column lists should not be empty")

        common_join_cols = []
        for (x, y) in zip(self.__column_names_self, self.__column_names_other):
            left_col = list(filter(lambda c: c.get_name() == x, self.__base_frame.columns()))[0]
            right_col = list(filter(lambda c: c.get_name() == y, self.__other_frame.columns()))[0]

            if left_col.get_type() != right_col.get_type():
                raise ValueError(
                    f"Trying to join on columns with different types -  Left Col: {left_col}, Right Col: {right_col}"
                )

            if x == y:
                common_join_cols.append(x)

        final_cols = (
                [x for x in left_cols if x not in common_join_cols] +
                [x for x in right_cols if x not in common_join_cols] +
                common_join_cols
        )

        if len(final_cols) != len(set(final_cols)):
            raise ValueError(
                "Found duplicate columns in joined frames (which are not join keys). "
                f"Columns -  Left Frame: {left_cols}, Right Frame: {right_cols}, Common Join Keys: {common_join_cols}"
            )

        if self.__join_type.lower() not in ('inner', 'left_outer', 'right_outer', 'leftouter', 'rightouter'):
            raise ValueError(
                f"Unknown join type - {self.__join_type}. Supported types are - INNER, LEFT_OUTER, RIGHT_OUTER"
            )

        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend.core.databse.sql_to_string import (
    SqlToStringGenerator,
    SqlToStringDbExtension,
)
from pylegend._typing import PyLegendSequence


__all__: PyLegendSequence[str] = [
    "PostgresSqlToStringGenerator"
]


class PostgresSqlToStringDbExtension(SqlToStringDbExtension):
    pass


class PostgresSqlToStringGenerator(SqlToStringGenerator):
    __sql_to_string_db_extension: SqlToStringDbExtension = PostgresSqlToStringDbExtension()

    @classmethod
    def database_type(cls) -> str:
        return "Postgres"

    @classmethod
    def create_sql_generator(cls) -> SqlToStringGenerator:
        return PostgresSqlToStringGenerator()

    def get_db_extension(self) -> SqlToStringDbExtension:
        return self.__sql_to_string_db_extension



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendList,
)
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SelectItem,
    SingleColumn,
    QualifiedName,
    QualifiedNameReference,
    AliasedRelation,
    Select,
    TableSubquery,
    Query
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig

__all__: PyLegendSequence[str] = [
    "create_sub_query",
    "copy_query",
    "extract_columns_for_subquery",
]


def create_sub_query(
        base_query: QuerySpecification,
        config: FrameToSqlConfig,
        alias: str,
        columns_to_retain: PyLegendOptional[PyLegendList[str]] = None
) -> QuerySpecification:
    query = copy_query(base_query)
    table_alias = config.sql_to_string_generator().get_db_extension().quote_identifier(alias)

    columns = extract_columns_for_subquery(query)
    outer_query_columns = columns_to_retain if columns_to_retain else columns
    unordered_select_items_with_index = [
        (
            outer_query_columns.index(x),
            SingleColumn(
                alias=x,
                expression=QualifiedNameReference(name=QualifiedName(parts=[table_alias, x]))
            )
        )
        for x in columns if x in outer_query_columns
    ]
    ordered_select_items: PyLegendList[SelectItem] = [
        y[1] for y in sorted(unordered_select_items_with_index, key=lambda x: x[0])
    ]

    return QuerySpecification(
        select=Select(
            selectItems=ordered_select_items,
            distinct=False
        ),
        from_=[
            AliasedRelation(
                relation=TableSubquery(query=Query(queryBody=query, limit=None, offset=None, orderBy=[])),
                alias=table_alias,
                columnNames=columns
            )
        ],
        where=None,
        groupBy=[],
        having=None,
        orderBy=[],
        limit=None,
        offset=None
    )


def copy_query(query: QuerySpecification) -> QuerySpecification:
    return QuerySpecification(
        select=copy_select(query.select),
        from_=query.from_,
        where=query.where,
        groupBy=query.groupBy,
        having=query.having,
        orderBy=query.orderBy,
        limit=query.limit,
        offset=query.offset
    )


def extract_columns_for_subquery(query: QuerySpecification) -> PyLegendList[str]:
    columns = []
    for col in query.select.selectItems:
        if not isinstance(col, SingleColumn):
            raise ValueError("Subquery creation not supported for queries "
                             "with columns other than SingleColumn")  # pragma: no cover
        if col.alias is None:
            raise ValueError("Subquery creation not supported for queries "
                             "with SingleColumns with missing alias")  # pragma: no cover
        columns.append(col.alias)
    return columns


def copy_select(select: Select) -> Select:
    return Select(
        distinct=select.distinct,
        selectItems=[s for s in select.selectItems]
    )



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
)
from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import LegendApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
    SelectItem,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_base_tds_frame import LegendApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "RenameColumnsFunction"
]


class RenameColumnsFunction(LegendApiAppliedFunction):
    __base_frame: LegendApiBaseTdsFrame
    __column_names: PyLegendList[str]
    __renamed_column_names: PyLegendList[str]

    @classmethod
    def name(cls) -> str:
        return "renameColumns"

    def __init__(
            self,
            base_frame: LegendApiBaseTdsFrame,
            column_names: PyLegendList[str],
            renamed_column_names: PyLegendList[str]
    ) -> None:
        self.__base_frame = base_frame
        self.__column_names = column_names
        self.__renamed_column_names = renamed_column_names

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)

        db_extension = config.sql_to_string_generator().get_db_extension()
        quoted_column_names_to_change = [db_extension.quote_identifier(s) for s in self.__column_names]
        quoted_renamed_column_names = [db_extension.quote_identifier(s) for s in self.__renamed_column_names]

        new_select_items: PyLegendList[SelectItem] = []
        for col in base_query.select.selectItems:
            if not isinstance(col, SingleColumn):
                raise ValueError("Rename columns operation not supported for queries "
                                 "with columns other than SingleColumn")  # pragma: no cover
            if col.alias is None:
                raise ValueError("Rename columns operation not supported for queries "
                                 "with SingleColumns with missing alias")  # pragma: no cover
            if col.alias in quoted_column_names_to_change:
                new_alias = quoted_renamed_column_names[quoted_column_names_to_change.index(col.alias)]
                new_select_items.append(SingleColumn(alias=new_alias, expression=col.expression))
            else:
                new_select_items.append(col)

        new_query = copy_query(base_query)
        new_query.select.selectItems = new_select_items
        return new_query

    def base_frame(self) -> LegendApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        new_columns = []
        for base_col in self.__base_frame.columns():
            if base_col.get_name() in self.__column_names:
                renamed_column_name = self.__renamed_column_names[self.__column_names.index(base_col.get_name())]
                new_columns.append(base_col.copy_with_changed_name(renamed_column_name))
            else:
                new_columns.append(base_col.copy())
        return new_columns

    def validate(self) -> bool:
        if len(self.__column_names) != len(self.__renamed_column_names):
            raise ValueError(
                "column_names list and renamed_column_names list should have same size when renaming columns.\n"
                f"column_names list - (Count: {len(self.__column_names)}) - {self.__column_names}\n"
                f"renamed_column_names_list - (Count: {len(self.__renamed_column_names)}) - {self.__renamed_column_names}\n"
            )

        if len(self.__column_names) != len(set(self.__column_names)):
            raise ValueError(
                "column_names list shouldn't have duplicates when renaming columns.\n"
                f"column_names list - (Count: {len(self.__column_names)}) - {self.__column_names}\n"
            )

        if len(self.__renamed_column_names) != len(set(self.__renamed_column_names)):
            raise ValueError(
                "renamed_column_names_list list shouldn't have duplicates when renaming columns.\n"
                f"renamed_column_names_list - (Count: {len(self.__renamed_column_names)}) - {self.__renamed_column_names}\n"
            )

        return True



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendTuple,
)
from pylegend.core.tds.legend_api.frames.legend_api_applied_function_tds_frame import LegendApiAppliedFunction
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SelectItem,
    SingleColumn,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.legend_api.frames.legend_api_base_tds_frame import LegendApiBaseTdsFrame
from pylegend.core.language import (
    TdsRow,
    AggregateSpecification,
    PyLegendPrimitive,
    create_primitive_collection,
    convert_literal_to_literal_expression,
)
from pylegend.core.tds.legend_api.frames.functions.function_helpers import tds_column_for_primitive


__all__: PyLegendSequence[str] = [
    "GroupByFunction"
]


class GroupByFunction(LegendApiAppliedFunction):
    __base_frame: LegendApiBaseTdsFrame
    __grouping_columns: PyLegendList[str]
    __aggregations: PyLegendList[AggregateSpecification]

    @classmethod
    def name(cls) -> str:
        return "group_by"

    def __init__(
            self,
            base_frame: LegendApiBaseTdsFrame,
            grouping_columns: PyLegendList[str],
            aggregations: PyLegendList[AggregateSpecification],
    ) -> None:
        self.__base_frame = base_frame
        self.__grouping_columns = grouping_columns
        self.__aggregations = aggregations

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        db_extension = config.sql_to_string_generator().get_db_extension()
        base_query = self.__base_frame.to_sql_query_object(config)

        should_create_sub_query = (len(base_query.groupBy) > 0) or base_query.select.distinct or \
                                  (base_query.offset is not None) or (base_query.limit is not None)

        columns_to_retain = [db_extension.quote_identifier(x) for x in self.__grouping_columns]
        if should_create_sub_query:
            new_query = create_sub_query(base_query, config, "root")
        else:
            new_query = copy_query(base_query)

        new_cols_with_index: PyLegendList[PyLegendTuple[int, 'SelectItem']] = []
        for col in new_query.select.selectItems:
            if not isinstance(col, SingleColumn):
                raise ValueError("Group By operation not supported for queries "
                                 "with columns other than SingleColumn")  # pragma: no cover
            if col.alias is None:
                raise ValueError("Group By operation not supported for queries "
                                 "with SingleColumns with missing alias")  # pragma: no cover
            if col.alias in columns_to_retain:
                new_cols_with_index.append((columns_to_retain.index(col.alias), col))

        new_select_items = [y[1] for y in sorted(new_cols_with_index, key=lambda x: x[0])]

        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        for agg in self.__aggregations:
            map_result = agg.get_map_fn()(tds_row)
            collection = create_primitive_collection(map_result)
            agg_result = agg.get_aggregate_fn()(collection)

            if isinstance(agg_result, (bool, int, float, str)):
                agg_sql_expr = convert_literal_to_literal_expression(agg_result).to_sql_expression(
                    {"frame": new_query},
                    config
                )
            else:
                agg_sql_expr = agg_result.to_sql_expression({"frame": new_query}, config)

            new_select_items.append(
                SingleColumn(alias=db_extension.quote_identifier(agg.get_name()), expression=agg_sql_expr)
            )

        new_query.select.selectItems = new_select_items
        new_query.groupBy = [
            (lambda x: x[c])(tds_row).to_sql_expression({"frame": new_query}, config) for c in self.__grouping_columns
        ]
        return new_query

    def base_frame(self) -> LegendApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        new_columns = []
        base_columns = self.__base_frame.columns()
        for c in self.__grouping_columns:
            for base_col in base_columns:
                if base_col.get_name() == c:
                    new_columns.append(base_col.copy())

        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        for agg in self.__aggregations:
            map_result = agg.get_map_fn()(tds_row)
            collection = create_primitive_collection(map_result)
            agg_result = agg.get_aggregate_fn()(collection)
            new_columns.append(tds_column_for_primitive(agg.get_name(), agg_result))

        return new_columns

    def validate(self) -> bool:
        base_columns = self.__base_frame.columns()
        for c in self.__grouping_columns:
            found_col = False
            for base_col in base_columns:
                if base_col.get_name() == c:
                    found_col = True
                    break
            if not found_col:
                raise ValueError(
                    f"Column - '{c}' in group by columns list doesn't exist in the current frame. "
                    f"Current frame columns: {[x.get_name() for x in base_columns]}"
                )

        agg_cols = [c.get_name() for c in self.__aggregations]
        new_cols = self.__grouping_columns + agg_cols

        if len(new_cols) == 0:
            raise ValueError("At-least one grouping column or aggregate specification must be provided "
                             "when using group_by function")

        if len(new_cols) != len(set(new_cols)):
            raise ValueError("Found duplicate column names in grouping columns and aggregation columns. "
                             f"Grouping columns - {self.__grouping_columns}, Aggregation columns - {agg_cols}")

        tds_row = TdsRow.from_tds_frame("frame", self.__base_frame)
        index = 0
        for agg in self.__aggregations:
            map_fn_copy = agg.get_map_fn()  # For MyPy
            if not isinstance(map_fn_copy, type(lambda x: 0)) or (map_fn_copy.__code__.co_argcount != 1):
                raise TypeError(
                    f"AggregateSpecification at index {index} (0-indexed) incompatible. "
                    "Map function should be a lambda which takes one argument (TDSRow)"
                )
            try:
                map_result = agg.get_map_fn()(tds_row)
            except Exception as e:
                raise RuntimeError(
                    f"AggregateSpecification at index {index} (0-indexed) incompatible. "
                    f"Error occurred while evaluating map function. Message: {str(e)}"
                ) from e

            if not isinstance(map_result, (int, float, bool, str, PyLegendPrimitive)):
                raise ValueError(
                    f"AggregateSpecification at index {index} (0-indexed) incompatible. "
                    f"Map function returns non-primitive - {str(type(map_result))}"
                )

            collection = create_primitive_collection(map_result)

            agg_fn_copy = agg.get_aggregate_fn()  # For MyPy
            if not isinstance(agg_fn_copy, type(lambda x: 0)) or (agg_fn_copy.__code__.co_argcount != 1):
                raise TypeError(
                    f"AggregateSpecification at index {index} (0-indexed) incompatible. "
                    "Aggregate function should be a lambda which takes one argument (primitive collection)"
                )
            try:
                agg_result = agg.get_aggregate_fn()(collection)
            except Exception as e:
                raise RuntimeError(
                    f"AggregateSpecification at index {index} (0-indexed) incompatible. "
                    f"Error occurred while evaluating aggregate function. Message: {str(e)}"
                ) from e

            if not isinstance(agg_result, PyLegendPrimitive):
                raise ValueError(
                    f"AggregateSpecification at index {index} (0-indexed) incompatible. "
                    f"Aggregate function returns non-primitive - {str(type(agg_result))}"
                )

            index += 1

        return True
